# coding: utf-8
require 'libmagellan'

require 'base64'
require 'json'
require 'active_support' 
require 'active_support/core_ext'
require 'signet/oauth_1/client'
require 'mqtt'


module Libmagellan

  class MQTT

    AUTH_BASE_URL = "mqtt://mqtt.magellanic-clouds.com".freeze
    MQTT_METHOD = "mqtt".freeze
    PROTOCOL = "mqtt".freeze

    LOG_LEVELS = {
      debug: Logger::DEBUG,
      info:  Logger::INFO,
      warn:  Logger::WARN,
      error: Logger::ERROR,
      fatal: Logger::FATAL,
    }
    DEFAULT_LOG_LEVEL = Logger::INFO

    COLOURS = {
      red:    "\e[31m",
      blue:   "\e[34m",
      green:  "\e[32m",
      yellow: "\e[33m",
      white:  "\e[37m",
    }

    OPTIONS = {
      host: {required: true},
      port: {required: true},
      consumer_key: {required: true},
      consumer_secret: {required: true},
      client_version: {required: true},
      secure: {default: false},
    }.freeze

    attr_accessor :host, :port, :consumer_key, :consumer_secret, :client_version
    attr_reader   :client
    attr_accessor :token

    # MQTT Client
    # @param [Hash] args MQTT Arguments
    # @option args [String] :host MQTT server host address
    # @option args [Integer] :port MQTT server port
    # @option args [String] :project Project name
    # @option args [String] :consumer_key MQTT authorization consumer-key
    # @option args [String] :consumer_key MQTT authorization consumer-secret
    def initialize(args={})
      args, errors = validate_args(args)
      show_errors(errors, initialize_usage()) if errors.present?

      @host = args.delete(:host)
      @port = args.delete(:port)
      @project = args.delete(:project)
      @consumer_key = args.delete(:consumer_key) || @project
      @consumer_secret = args.delete(:consumer_secret)
      @client_version = args.delete(:client_version)
      @secure = args.delete(:secure) || false
      @skip_verify = args[:skip_verify] || false

      @token = nil
      @client = nil
    end


    # Publish message to MQTT server.
    # @param [String] topic Topic name
    # @param [String] payload Payload
    def publish(topic, payload, retain=false, qos=0)
      payload = payload.to_json unless payload.is_a?(::String)
      with_connection do |c|
        c.publish(topic, payload, retain, qos)
        log("[PUBLISH] #{topic}: #{payload}", Logger::DEBUG)
      end
    end
    alias :pub :publish

    # Subscribe topic(s)
    # @param [String|Array|Hash] topics Subscribe topics
    def subscribe(*topics, &block)
      with_connection do |c|
        c.subscribe(*topics)
        topics.each {|t| log("[SUBSCRIBE] #{t}", Logger::DEBUG) }
      end
      if block_given?
        get(nil, &block)
      end
    end
    alias :sub :subscribe

    # Unsubscribe topic(s)
    # @param [String|Array|Hash] topics Subscribe topics
    def unsubscribe(*topics)
      with_connection do |c|
        c.unsubscribe(*topics)
        topics.each {|t| log("[UNSUBSCRIBE] #{t}", Logger::DEBUG) }
      end
    end
    alias :unsub :unsubscribe

    # Get message to me
    # @param [String] topic Topic name
    # @return [Array] topic and payload
    def get(topic=nil)
      with_connection do |c|
        if block_given?
          c.get(topic) do |topic_, payload_|
            yield(topic_, payload_)
          end
        else
          # return topic, payload
          c.get(topic)
        end
      end
    end

    # Get message to me
    # @param [String] topic Topic name
    # @return [Mqtt::Packet] MQTT packet object
    def get_packet(topic=nil)
      with_connection do |c|
        if block_given?
          c.get_packet(topic) do |packet|
            yield(packet)
          end
        else
          # return topic, payload
          c.get_packet(topic)
        end
      end
    end


    # MQTT Disconnect
    def disconnect
      connection.disconnect if connected?
    end

    # MQTT Will
    def set_will(topic, payload, retain=false, qos=0)
      @will_topic = topic
      @will_payload = payload
      @will_qos = qos
      @will_retain = retain
      @with_will = true
    end

    # @return [Logger]
    def logger
      @logger_
    end

    def logger=(logger_)
      @logger_ = logger_
    end



    private


    # Send command to MQTT server.
    # Connect MQTT server if it is not has connection.
    def with_connection
      yield(connection()) if block_given?
    end

    # Connect MQTT server
    # @return [MQTT::Client] MQTT Client
    def connection
      username = @project || @consumer_key
      if connected?
        return @client
      else
        if @token.present?
          # トークンによる認証
          connect(username, @token)
        else
          # OAuthによる認証
          auth = generate_oauth_header(AUTH_BASE_URL, @consumer_key, @consumer_secret)
          headers = "Authorization: #{auth}"
          headers << ";Client-Version: #{@client_version}" if @client_version.present?
          password = Base64.urlsafe_encode64(headers)
          connect(username, password)
        end
      end
    end

    # Connect to MQTT Server
    # @param [String] username Username
    # @param [String] password Password
    # @return [MQTT::Client] MQTT Client
    def connect(username, password)
      mqtt_host = "#{@host}:#{@port.to_s}"
      uri = "#{PROTOCOL}://#{username}:#{password}@#{mqtt_host}"
      @client = ::MQTT::Client.new(uri)
      if @with_will
        @client.set_will(@will_topic, @will_payload, @will_retain, @will_qos)
      end
      if @secure
        @client.ssl = true
        @client.ssl_context.verify_mode = OpenSSL::SSL::VERIFY_NONE if @skip_verify
      end
      @client.connect
      log("[CONNECTED] #{host}:#{port}", Logger::DEBUG)
      return @client
    end


    # Check has connection.
    # @return [Boolean]
    def connected?
      @client.present? and @client.connected?
    end

    # Generate OAuth header
    # @param [String] uri Auth URI
    # @param [String] consumer_key Auth consumer-key
    # @param [String] consumer_secret Auth consumer-secret
    # @param [Hash] options Auth options
    def generate_oauth_header(uri, consumer_key, consumer_secret, options={})
      init_opt = {client_credential_key: consumer_key, client_credential_secret: consumer_secret}
      options[:uri] = uri if options[:uri].nil? or options[:uri].empty?
      options[:method] = MQTT_METHOD
      client = ::Signet::OAuth1::Client.new(init_opt)
      client.two_legged = true
      client, options = yield(client, options) if block_given?
      req = client.generate_authenticated_request(options)
      req["Authorization"]
    end


    # Validate arguments.
    # @see initialize
    def validate_args(args)
      errors = []
      args = args.symbolize_keys
      OPTIONS.each do |key, val|
        if val[:required]
          errors << {field: key, message: "required param"} if args[key].blank?
        end
        if val[:default] and not args.key?(key)
          args[key] = val[:default]
        end
      end
      return args, errors
    end

    # Logging
    # @param [String] message Logging message
    # @param [Symbol|String] type Logging type
    # @param [Symbol] colour Message color
    def log(message="", lv=DEFAULT_LOG_LEVEL, colour=nil)
      color = COLOURS[colour.to_s.downcase.to_sym]
      m = ""
      m << "#{color}" if color.present?
      m << "#{message}"
      m << "\e[0m" if color.present?
      logger.log(lv, m) if logger
    end


    # Show errors and exit(0)
    # @param [Array] errors Error messages
    # @param [String] message Extra message
    def show_errors(errors, message="")
      errors.each do |e|
        log("#{e[:field]}: #{e[:message]}", Logger::ERROR, :red)
      end
      log(message, Logger::INFO, :white) if message.present?
    end

    def initialize_usage
      <<__USAGE__

usage:

c = Libmagellan::MQTT.new(host: "127.0.0.1", host: 1883, consumer_key: "groovenauts.app1", consumer_secret: "test", client_version: "0.0.1", log_level: :debug, commandline: true)

__USAGE__
    end

    class MqttError < StandardError
    end

  end
end
